/*
	cluster_mapping.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2007,8 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include "mapping.h"

void print_to_gff(FILE* gfffile,char* sort_target,int cluster_start,int cluster_end,double cluster_expression,int count,char strand,int single_mapper,int id);
void print_to_bed(FILE* bedfile,char* sort_target,int cluster_start,int cluster_end,double cluster_expression,char strand,int single_mapper,int id);

void print_to_gff(FILE* gfffile,char* sort_target,int cluster_start,int cluster_end,double cluster_expression,int count,char strand,int single_mapper,int id)
{
	if(count > 1){
		fprintf(gfffile,"%s\tNexalign\t.\t%d\t%d\t%f\t%c\t.\tID=Cluster%d;sequences=%d;single=%d\n",sort_target,cluster_start,cluster_end,cluster_expression,strand,id,count,single_mapper);
	}else{
		fprintf(gfffile,"%s\tNexalign\t.\t%d\t%d\t%f\t%c\t.\tID=Cluster%d;sequences=%d;single=%d\n",sort_target,cluster_start,cluster_end,cluster_expression,strand,id,count,single_mapper);
	}
}

void print_to_bed(FILE* bedfile,char* sort_target,int cluster_start,int cluster_end,double cluster_expression,char strand,int single_mapper,int id)
{
	if(strand == '+'){
		//fprintf(bedfile,"%s\t%d\t%d\tCluster%d\t%f\t%c\t%d\t%d\t255,0,0\n",sort_target,cluster_start,cluster_end,id,cluster_expression,strand,cluster_start,cluster_end);
		fprintf(bedfile,"%s\t%d\t%d\t%f\n",sort_target,cluster_start,cluster_end,cluster_expression);
	}else{
		//fprintf(bedfile,"%s\t%d\t%d\tCluster%d\t%f\t%c\t%d\t%d\t0,0,255\n",sort_target,cluster_start,cluster_end,id,cluster_expression,strand,cluster_start,cluster_end);
		fprintf(bedfile,"%s\t%d\t%d\t%f\n",sort_target,cluster_start,cluster_end,cluster_expression);
	}
}



void get_mapped_expression(struct parameters* param)
{
	FILE *file = 0;
	char test1[MAX_LINE];
	char test2[MAX_LINE];
	char test3[MAX_LINE];
	char test4[MAX_LINE];
	float tag_count = 0.0;
	int start,stop;
	int map_num = 0;
	
	
	if (!(file = fopen(param->outfile, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	
	while(fscanf(file, "%s%s%s%d%d%s%f%d", test1,test2,test3,&start,&stop,test4,&tag_count,&map_num)!= EOF){
		param->mapped_expression += (double)tag_count/(double)map_num;
	}

	fclose(file);
}





void cluster_mapping(struct parameters* param)
{
	struct mapping** map = 0;
	FILE *file = 0;
	FILE *foutfile = 0;
	FILE *gfffile = 0;
	FILE *bedfile_plus = 0;
	FILE *bedfile_minus = 0;
	char *time_s = 0;
	char* sort_target = 0;
	
	char test1[MAX_LINE];
	
	char test2[MAX_LINE];
	char test3[MAX_LINE];
	char test4[MAX_LINE];
	float tag_count = 0.0;
	
	
	double cluster_expression = 0.0;
	int map_num = 0;
	int start,stop,i;
	int count = 0;
	int window = param->cluster;
	int cluster_num =1;
	int single_mapper = 0;
	int cluster_start = -1000;
	int cluster_end = -1000;
	//double total = 0.0 ;
	
	time_s = get_time(time_s);
	fprintf(stderr,"%s	Clustering Mapping\n",time_s);
	
	
	if (!(file = fopen(param->outfile, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	
	if ((foutfile = fopen(param->outfile_tmp, "w")) == NULL){
		fprintf(stderr,"can't open output\n");
		exit(-1);
	}
	
	if(param->gff){
		if((gfffile = fopen(param->gff, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		fprintf(gfffile,"##gff-version 3\n");
		fprintf(gfffile,"#source-version %s\n",VERSION);
		fprintf(gfffile,"#date %s\n",time_s);
	}
	if(param->bed){
		for(i = 0; i < strlen(param->bed);i++){
			test1[i] = param->bed[i];
		}
		test1[i] = '_';
		test1[i+1] = '+';
		test1[i+2] = '.';
		test1[i+3] = 'b';
		test1[i+4] = 'e';
		test1[i+5] = 'd';
		test1[i+6] = 0;
		
		if((bedfile_plus = fopen(test1, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		for(i = 0; i < strlen(param->bed);i++){
			test1[i] = param->bed[i];
		}
		test1[i] = '_';
		test1[i+1] = '-';
		test1[i+2] = '.';
		test1[i+3] = 'b';
		test1[i+4] = 'e';
		test1[i+5] = 'd';
		test1[i+6] = 0;
		
		
		if((bedfile_minus = fopen(test1, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
		fprintf(bedfile_plus,"track type=bedGraph name=\"%s +\" description=\"%s +\" visibility=full color=0,0,255 altColor=0,100,200 priority=20\n",param->description,param->description);
		fprintf(bedfile_minus,"track type=bedGraph name=\"%s -\" description=\"%s -\" visibility=full color=0,0,255 altColor=0,100,200 priority=20\n",param->description,param->description);
		
	}
	

	while(fscanf(file, "%s%s%s%d%d%s%f%d", test1,test2,test3,&start,&stop,test4,&tag_count,&map_num)!= EOF){
		if(test3[0] == '+'){
			if(!sort_target){
				sort_target = strdup(test2);
			}
			if(!strcmp(sort_target,test2)){
				if(cluster_start - window <= start && start < cluster_end + window){					
					map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
					map[count] = (struct mapping *)malloc(sizeof(struct mapping));
					map[count]->qname = strdup(test1);
					map[count]->tname = strdup(test2);
					map[count]->strand = strdup(test3);
					map[count]->edit = strdup(test4);
					map[count]->start = start;
					map[count]->stop = stop;
					map[count]->tag_count = tag_count;
					map[count]->map_num = map_num;
					cluster_expression +=  (double) tag_count / (double) map_num;
					count++;
					if(map_num == 1){
						single_mapper++;
					}
					if(stop > cluster_end){
						cluster_end = stop;
					}
				}else{
					if(cluster_start != -1000){
						//total += cluster_expression;
						cluster_expression = cluster_expression/param->mapped_expression*1000000;
						fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'+',cluster_start,cluster_end,count,cluster_expression);
						if(param->gff){
							print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'+',single_mapper,cluster_num);
						//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t+\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
						}
						
						if(param->bed){
							print_to_bed(bedfile_plus,sort_target,cluster_start,cluster_end,cluster_expression,'+',single_mapper,cluster_num);
						}
						
						//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'+',cluster_start,cluster_end);
						for(i = 0; i < count; i++) {
							fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
						}	
						for(i = 0; i < count;i++) {
							free(map[i]->qname);
							free(map[i]->tname);
							free(map[i]->strand);
							free(map[i]->edit);
							free(map[i]);
						}
						free(map);
						map = 0;
						count = 0;
						single_mapper = 0;
						cluster_expression = 0;
						cluster_num++;
					}
					cluster_start = start;
					cluster_end = stop;	
					
					map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
					map[count] = (struct mapping *)malloc(sizeof(struct mapping));
					map[count]->qname = strdup(test1);
					map[count]->tname = strdup(test2);
					map[count]->strand = strdup(test3);
					map[count]->edit = strdup(test4);
					map[count]->start = start;
					map[count]->stop = stop;
					map[count]->tag_count = tag_count;
					map[count]->map_num = map_num;
					cluster_expression +=  (double) tag_count / (double) map_num;
					count++;
					if(map_num == 1){
						single_mapper++;
					}
				}
			}else{
				if(cluster_start != -1000){
					//total += cluster_expression;
					cluster_expression = cluster_expression/param->mapped_expression*1000000;
					fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'+',cluster_start,cluster_end,count,cluster_expression);
					if(param->gff){
						print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'+',single_mapper,cluster_num);
					//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t+\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
					}
					
					if(param->bed){
						print_to_bed(bedfile_plus,sort_target,cluster_start,cluster_end,cluster_expression,'+',single_mapper,cluster_num);
					}
					//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'+',cluster_start,cluster_end);
					for(i = 0; i < count; i++) {
						fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
					}	
					for(i = 0; i < count;i++) {
						free(map[i]->qname);
						free(map[i]->tname);
						free(map[i]->strand);
						free(map[i]->edit);
						free(map[i]);
					}
					free(map);
					map = 0;
					count = 0;
					single_mapper = 0;
					cluster_expression = 0;
					cluster_num++;
				}
				free(sort_target);
				sort_target = 0;
				sort_target = strdup(test2);
				cluster_start = start;
				cluster_end = stop;	
					
				map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
				map[count] = (struct mapping *)malloc(sizeof(struct mapping));
				map[count]->qname = strdup(test1);
				map[count]->tname = strdup(test2);
				map[count]->strand = strdup(test3);
				map[count]->edit = strdup(test4);
				map[count]->start = start;
				map[count]->stop = stop;
				map[count]->tag_count = tag_count;
				map[count]->map_num = map_num;
				cluster_expression +=  (double) tag_count / (double) map_num;
				count++;
				if(map_num == 1){
					single_mapper++;
				}
			}
		}
	}
	if(count){
		//total += cluster_expression;
		cluster_expression = cluster_expression/param->mapped_expression*1000000;
		fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'+',cluster_start,cluster_end,count,cluster_expression);
		if(param->gff){
			//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t+\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
			print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'+',single_mapper,cluster_num);
		}
		
		if(param->bed){
			print_to_bed(bedfile_plus,sort_target,cluster_start,cluster_end,cluster_expression,'+',single_mapper,cluster_num);
		}
		
		//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'+',cluster_start,cluster_end);
		for(i = 0; i < count; i++) {
			fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
		}	
		for(i = 0; i < count;i++) {
			free(map[i]->qname);
			free(map[i]->tname);
			free(map[i]->strand);
			free(map[i]->edit);
			free(map[i]);
		}
		free(map);
		
		cluster_num++;
	}
	free(sort_target);
	sort_target = 0;
	
	map = 0;
	count =0;
	single_mapper = 0;
	cluster_expression = 0;
	cluster_start = -1000;
	cluster_end = -1000;
	rewind(file);
	
	
	while(fscanf(file, "%s%s%s%d%d%s%f%d", test1,test2,test3,&start,&stop,test4,&tag_count,&map_num)!= EOF){
		if(test3[0] == '-'){
			if(!sort_target){
				sort_target = strdup(test2);
			}
			if(!strcmp(sort_target,test2)){
				if(cluster_start - window <= start && start < cluster_end + window){					
					map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
					map[count] = (struct mapping *)malloc(sizeof(struct mapping));
					map[count]->qname = strdup(test1);
					map[count]->tname = strdup(test2);
					map[count]->strand = strdup(test3);
					map[count]->edit = strdup(test4);
					map[count]->start = start;
					map[count]->stop = stop;
					map[count]->tag_count = tag_count;
					map[count]->map_num = map_num;
					cluster_expression +=  (double) tag_count / (double) map_num;
					count++;
					if(map_num == 1){
						single_mapper++;
					}
					if(stop > cluster_end){
						cluster_end = stop;
					}
				}else{
					if(cluster_start != -1000){
						//total += cluster_expression;
						cluster_expression = cluster_expression/param->mapped_expression*1000000;
						fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'-',cluster_start,cluster_end,count,cluster_expression);
						if(param->gff){
							print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'-',single_mapper,cluster_num);
							//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t-\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
						}
						
						if(param->bed){
							print_to_bed(bedfile_minus,sort_target,cluster_start,cluster_end,cluster_expression,'-',single_mapper,cluster_num);
						}
						
						
						//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'-',cluster_start,cluster_end);
						for(i = 0; i < count; i++) {
							fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
						}	
						for(i = 0; i < count;i++) {
							free(map[i]->qname);
							free(map[i]->tname);
							free(map[i]->strand);
							free(map[i]->edit);
							free(map[i]);
						}
						free(map);
						map = 0;
						count = 0;
						single_mapper = 0;
						cluster_expression = 0;
						cluster_num++;
					}
					cluster_start = start;
					cluster_end = stop;	
					
					map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
					map[count] = (struct mapping *)malloc(sizeof(struct mapping));
					map[count]->qname = strdup(test1);
					map[count]->tname = strdup(test2);
					map[count]->strand = strdup(test3);
					map[count]->edit = strdup(test4);
					map[count]->start = start;
					map[count]->stop = stop;
					map[count]->tag_count = tag_count;
					map[count]->map_num = map_num;
					cluster_expression +=  (double) tag_count / (double) map_num;
					count++;
					if(map_num == 1){
						single_mapper++;
					}
				}
			}else{
				if(cluster_start != -1000){
					//total += cluster_expression;
					cluster_expression = cluster_expression/param->mapped_expression*1000000;
					fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'-',cluster_start,cluster_end,count,cluster_expression);
					if(param->gff){
						print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'-',single_mapper,cluster_num);
					//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t-\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
					}
					
					if(param->bed){
						print_to_bed(bedfile_minus,sort_target,cluster_start,cluster_end,cluster_expression,'-',single_mapper,cluster_num);
					}
					
					//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'-',cluster_start,cluster_end);
					for(i = 0; i < count; i++) {
						fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
					}	
					for(i = 0; i < count;i++) {
						free(map[i]->qname);
						free(map[i]->tname);
						free(map[i]->strand);
						free(map[i]->edit);
						free(map[i]);
					}
					free(map);
					map = 0;
					count = 0;
					single_mapper = 0;
					cluster_expression = 0;
					cluster_num++;
				}
				free(sort_target);
				sort_target = 0;
				sort_target = strdup(test2);
				cluster_start = start;
				cluster_end = stop;	
					
				map = (struct mapping**)realloc(map, (count + 1) * sizeof(struct mapping*));
				map[count] = (struct mapping *)malloc(sizeof(struct mapping));
				map[count]->qname = strdup(test1);
				map[count]->tname = strdup(test2);
				map[count]->strand = strdup(test3);
				map[count]->edit = strdup(test4);
				map[count]->start = start;
				map[count]->stop = stop;
				map[count]->tag_count = tag_count;
				map[count]->map_num = map_num;
				cluster_expression +=  (double) tag_count / (double) map_num;
				count++;
				if(map_num == 1){
					single_mapper++;
				}
			}
		}
	}
	if(count){
		//total += cluster_expression;
		cluster_expression = cluster_expression/param->mapped_expression*1000000;
		fprintf(foutfile,"Cluster%d:%s_%c_%d-%d_%d_%f\n",cluster_num,sort_target,'-',cluster_start,cluster_end,count,cluster_expression);
		if(param->gff){
			print_to_gff(gfffile,sort_target,cluster_start,cluster_end,cluster_expression,count,'-',single_mapper,cluster_num);
			//fprintf(gfffile,"%s\tNexalign\tTag_cluster\t%d\t%d\t%f\t-\t.\t%d sequences\n",sort_target,cluster_start,cluster_end,cluster_expression,count);
		}
		
		if(param->bed){
			print_to_bed(bedfile_minus,sort_target,cluster_start,cluster_end,cluster_expression,'-',single_mapper,cluster_num);
		}
		
		//fprintf(stderr,"Cluster%d:%s_%c_%d-%d\n",cluster_num,sort_target,'-',cluster_start,cluster_end);
		for(i = 0; i < count; i++) {
			fprintf(foutfile,"%s	%s	%s	%d	%d	%s	%f	%d\n",map[i]->qname,map[i]->tname,map[i]->strand,map[i]->start,map[i]->stop,map[i]->edit,map[i]->tag_count,map[i]->map_num);
		}	
		for(i = 0; i < count;i++) {
			free(map[i]->qname);
			free(map[i]->tname);
			free(map[i]->strand);
			free(map[i]->edit);
			free(map[i]);
		}
		free(map);
		cluster_num++;
	}
	
	free(sort_target);
	sort_target = 0;
	
	if(param->gff){
		fclose(gfffile);
	}
	if(param->bed){
		fclose(bedfile_plus);
		fclose(bedfile_minus);
	}
	
	
	fclose(file);
	fclose(foutfile);	
	
	//fprintf(stderr,"%f\n%f\n",total,param->mapped_expression);
	
	
	if (!(file = fopen(param->outfile_tmp, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->outfile);
	}
	
	if(param->input_clustering){
		if ((foutfile = fopen(param->fasta_file, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}else{
		if ((foutfile = fopen(param->outfile, "w")) == NULL){
			fprintf(stderr,"can't open output\n");
			exit(-1);
		}
	}
	
	while(fgets(test1, MAX_LINE, file)){
		fprintf(foutfile,"%s",test1);
	}
	fclose(foutfile);
	fclose(file);
	
	if(param->input_clustering){
		free(param->fasta_file);
	}
	
	if(!param->quiet_flag){
		time_s = get_time(time_s);
		fprintf(stderr,"%s	Done (found %d clusters)\n",time_s,cluster_num-1);
	}
}






