/*
	input.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"
#include <ctype.h>

struct fasta* convert_fasta_cs(struct fasta* aln);

void set_hash_size(struct parameters* param)
{
	FILE *file = 0;
	int i;
	
	if (!(file = fopen(param->fasta_file, "r" ))){
		fprintf(stderr,"Cannot open file '%s'\n",param->fasta_file);
		exit(-1);
	}
	if (fseek(file,0,SEEK_END) != 0){
		(void)fprintf(stderr, "ERROR: fseek failed\n");
		(void)exit(EXIT_FAILURE);
	}
	i= ftell (file);
	if (fseek(file,0,SEEK_START) != 0){
		(void)fprintf(stderr, "ERROR: fseek failed\n");
		(void)exit(EXIT_FAILURE);
	}
	//fprintf(stderr,"%s	%d	%d	%d\n",param->fasta_file,i,i/30,param->query_num_limit);
	if((int)((i / 30) + 1) < param->query_num_limit){
		param->query_num_limit = (int)((i / 30)+1);
	}
	fclose(file);
	//exit(0);
	//return 0;
}



struct fasta* get_fasta(struct fasta* p,char *infile,int solid)
{
	p = (struct fasta*) malloc(sizeof(struct fasta));
	assert(p != 0);
	p->string = 0;
	p->string =  get_input_into_string(p->string,infile);
	if(!p->string){
		fprintf(stderr,"Analysing input %s ... nothing\n",infile);
		exit(-1);
	}
	if(solid){
		p = convert_fasta_cs(p);
	}else{
		p = read_fasta(p);
	}
	return p;
}

char* get_input_into_string(char* string,char* infile)
{
	int i = 0;   
	int string_length = 2;
	char c = 0;    
	
	/*
		// added these items Jody Hey 11/29/07
	int maxl, tempmaxl, sl;
	char* cc;
	int j;
	*/
	FILE *file = 0;
	if(infile){
		if (!(file = fopen( infile, "r" ))){
			return 0;
 			fprintf(stderr,"Cannot open file '%s'\n", infile);
			exit(-1);
		}
		if (fseek(file,0,SEEK_END) != 0){
			(void)fprintf(stderr, "ERROR: fseek failed\n");
			(void)exit(EXIT_FAILURE);
		}
		i= ftell (file);
		if (fseek(file,0,SEEK_START) != 0){
			(void)fprintf(stderr, "ERROR: fseek failed\n");
			(void)exit(EXIT_FAILURE);
		}
		if(!string){
			string = malloc ((i+1+18)* sizeof(char));
			assert(string != 0);
		}
		fread(string,sizeof(char), i, file);
		string[i] = 0;
		fclose(file);
		/*
		// added this section of code, Jody Hey 11/29/07
// gets the longest line length,  so that we can use fgets later, then uses fgets to get a string of the file contents
		maxl = tempmaxl = 0;
		sl = strlen(string) - 1;
		for  (j = 0; j <= sl; j++)
			if (string[j]=='\n' || j == sl){
			if (maxl < (j - tempmaxl))
				maxl = tempmaxl;
			tempmaxl = j;
			}
			maxl++;
			string[0] = 0;
			if (!(file = fopen( infile, "r" ))){
				fprintf(stderr,"Cannot open file '%s'\n", infile);
				exit(1);
			}
			j = 0;
			cc = string;
			while (fgets(cc, maxl,file))
				cc = &string[strlen(string)]; 
			fclose(file);
// end of added section */
	}else{
		//fprintf(stderr,"from STDIN\n");
		if (!isatty(0)){
			string = malloc(sizeof(char*)*string_length);
			assert(string != 0);
			while (!feof (stdin)){
				c = getc(stdin);
				if (i == string_length){
					string_length <<=1;
					string = realloc(string,sizeof(char)*(string_length+18));
				}
				string[i] = c;
				i++;
			}
			string[i-1] = 0;
			//fprintf(stderr,"STRING:%s\n",string);
		}else{
			return 0;
		}
	}
	return string;
}

struct fasta* convert_fasta_cs(struct fasta* aln)
{
	char solid[4][4] = {{'0','1','2','3'},{'1','0','3','2'},{'2','3','0','1'},{'3','2','1','0'}};
	int before = -1;
	int now = -1;
	int c = 0;
	int n = 0;
	int i = 0;
	int j = 0;
	int len = 0;
	int stop = 0;
	int nbytes;
	nbytes = strlen(aln->string);
	aln->numseq = 0;
	aln->max_len = -1;
	//aln->org_seq = 0;
	stop = 0;
	//aln->org_seq = malloc(sizeof(char)*nbytes);
	//assert(aln->org_seq != 0);
	//count filenames....
	for (i =0;i < nbytes;i++){
		if (aln->string[i] == '>'&& stop == 0){
			aln->numseq++;
			stop = 1;
		}else if (aln->string[i] == '\n'){
			stop = 0;
		}else if(aln->string[i] == '\r'){
			aln->string[i] = '\n';
		}
	}
	
	aln->suffix = 0;

	aln->sn = malloc(sizeof(char*)*aln->numseq);
	assert(aln->sn != 0);

	//aln->c = 0;

	
	aln->s_index = malloc(sizeof(int)*(aln->numseq+1));
	assert(aln->s_index != 0);
	for (i =0;i < nbytes;i++){
		if (aln->string[i] == '>'){
			if(aln->max_len < len){
				aln->max_len = len;
			}
			len = 0;
			j = i+1;
			while(aln->string[j] != '\n'){
			//	fprintf(stderr,"%c",aln->string[j]);
				j++;
			}
			//fprintf(stderr,"	%d\n",j-i);
			aln->sn[c] = malloc(sizeof(char)*(j-i));
			assert(aln->sn[c]  != 0);
			aln->s_index[c] = n;
			j = i+1;
			len = 0;
			while(aln->string[j] != '\n'){
				aln->sn[c][len] = aln->string[j];
				len++;
				j++;
			}
			aln->sn[c][len] = 0;
			aln->string[n] = 'X';
			//aln->org_seq[n] = 'X';
			before = -1;
			n++;
			i+= j-i;
			c++;
		}else if (aln->string[i] != '\n' && aln->string[i] != 0 ){
			if(isalpha((int)aln->string[i])){
				//aln->org_seq[n] = toupper(aln->string[i]);
				switch(toupper(aln->string[i])){
					case 'A':
						now = 0;
						break;
					case 'C':
						now = 1;
						break;
					case 'G':
						now = 2;
						break;
					case 'T':
						now = 3;
						break;
					default:
						now = -1;
						break;
				}
				if(before == -1 || now == -1){
					aln->string[n] = toupper(aln->string[i]);
				}else{
					aln->string[n] = solid[before][now];
				}
				
				before = now;
				n++;
				len++;
			}
		}
	}
	aln->s_index[c] = n;
	aln->string[n] = 'X';
	//aln->org_seq[n] = 'X';
	aln->string[n+1] = 0;
	//aln->org_seq[n+1] = 0;
	aln->string_len = n+1;
	
	aln->string = (char*) realloc (aln->string, (n+2) * sizeof(char));
	//aln->org_seq =(char*) realloc (aln->org_seq, (n+2) * sizeof(char));
	//fprintf(stderr,"%s\n%s\n",aln->string,aln->org_seq);

	return aln;
}

struct parameters* read_list(struct parameters* param,char* string)
{
	int nbytes;
	int start = 0;
	int i,j,c;
	param->num_types = 0;
	nbytes = strlen(string);
	for (i =0;i < nbytes;i++){
		if (string[i] == '\n'){
			param->num_types++;
		}
	}
	param->types = malloc(sizeof(char*) * (param->num_types+1));
	assert(param->types != 0);
	for(i = 0; i < param->num_types+1;i++){
		param->types[i] = 0;
	}
	start = 0;
	c = 0;
	for (i = 0;i < nbytes;i++){
		if(string[i] == '\n'){
			if(i > start){
				param->types[c] = malloc(sizeof(char) *((i - start)+1));
				assert(param->types[c] != 0);
				for (j = start; j < i;j++){
					param->types[c][j-start] = string[j];
				}
				param->types[c][i-start] = 0; 
				c++;
			}
			start = i+1;
		}
	}
	if(start != i){
		if(i > start){
		param->types[c] = malloc(sizeof(char) *((i - start)+1));
		assert(param->types[c] != 0);
		for (j = start; j < i;j++){
			param->types[c][j-start] = string[j];
		}
		param->types[c][i-start] = 0; 
		param->num_types++;
		}
	}
	for(i = 0; i < param->num_types+1;i++){
		if(param->types[i] == 0){
			param->num_types = i;
		}
	}
	return param;
}


struct fasta* read_fasta(struct fasta* aln)
{
	int c = 0;
	int n = 0;
	int i = 0;
	int j = 0;
	int len = 0;
	int stop = 0;
	int nbytes;

	nbytes = strlen(aln->string);
	aln->numseq = 0;
	aln->max_len = -1;
	//aln->org_seq = 0;
	stop = 0;
	
	//count filenames....
	for (i =0;i < nbytes;i++){
		if (aln->string[i] == '>'&& stop == 0){
			aln->numseq++;
			stop = 1;
		}else if (aln->string[i] == '\n'){
			stop = 0;
		}else if(aln->string[i] == '\r'){
			aln->string[i] = '\n';
		}
	}
	
	aln->suffix = 0;
	aln->sn = malloc(sizeof(char*)*aln->numseq);
	assert(aln->sn  !=0);
	//aln->c = 0;

	aln->s_index = malloc(sizeof(int)*(aln->numseq+1));
	assert(aln->s_index != 0);
	for (i =0;i < nbytes;i++){
		if (aln->string[i] == '>'){
			if(aln->max_len < len){
				aln->max_len = len;
			}
			
			len = 0;
			j = i+1;
			while(aln->string[j] != '\n'){
			//	fprintf(stderr,"%c",aln->string[j]);
				j++;
			}
			//fprintf(stderr,"	%d\n",j-i);
			aln->sn[c] = malloc(sizeof(char)*(j-i));
			assert(aln->sn[c] != 0);
			aln->s_index[c] = n;
			j = i+1;
			len = 0;
			while(aln->string[j] != '\n'){
				if(isspace((int)aln->string[j])){
					aln->sn[c][len] = '_';
				}else{ 
					aln->sn[c][len] = aln->string[j];
				}
				len++;
				j++;
			}
			aln->sn[c][len] = 0;
			aln->string[n] = 'X';
			n++;
			i+= j-i;
			c++;
		}else if(isalnum((int)aln->string[i])){// if (aln->string[i] != '\n' && aln->string[i] != 0 ){
			aln->string[n] =  toupper(aln->string[i]);
			n++;
			len++;
		}
		//}
	}

	
	aln->s_index[c] = n;
	aln->string[n] = 'X';
	aln->string[n+1] = 0;

	aln->string_len = n+1;
	return aln;
}

void read_types_list(struct parameters* param)
{
	FILE *file;
	int i = 0;
	if (!(file = fopen(param->type_file, "r" ))){
		fprintf(stderr,"Cannot open query file '%s'\n", param->fasta_file);
		exit(-1);
	}
	param->types = malloc(sizeof(char*)*32);
	assert(param->types !=0);
	param->type_colors = malloc(sizeof(char*)*32);
	assert(param->type_colors != 0);
	for(i = 0; i < 32;i++){
		param->types[i] = malloc(sizeof(char)*MAX_LINE);
		param->type_colors[i] =  malloc(sizeof(char)*MAX_LINE);
		param->types[i][0] = 0;
		param->type_colors[i][0] = 0;
		
	}
	
	i = 0;
	while(!feof(file)){
		fscanf(file,"%s %s",param->types[i],param->type_colors[i]);
		if(!strlen(param->types[i]) || !strlen(param->type_colors[i])){
			break;
		}
		i++;
	}
	fclose(file);
	param->num_types = i;
	for(i = 0;i < param->num_types;i++){
		param->types[i] = realloc(param->types[i],sizeof(char)*(strlen(param->types[i])+1));
		param->type_colors[i] = realloc(param->type_colors[i],sizeof(char)*(strlen(param->type_colors[i])+1));
	}
	/*for(i = 0;i < param->num_types;i++){
		fprintf(stderr,"%d	%s	%s\n",i,param->types[i],param->type_colors[i]);
	}
	exit(0);*/
}

struct query_table* read_query_fasta_file_with_name(struct query_table* qt,struct parameters* param)
{
	FILE *file;
	char line[MAX_LINE];
	char tmp_seq[MAX_LINE];
	char tmp_name[MAX_LINE];
	int i;
	int len = 0;
	int sn_len = 0;
	double count = 0;
	int numseq = 0;
	
	if(!param->fasta_file){
		file = stdin;
		if (isatty(0)){
			//fprintf(stderr,"No STDIN..\n");
			return qt;
		}
	}else{
		if (!(file = fopen(param->fasta_file, "r" ))){
			fprintf(stderr,"Cannot open query file '%s'\n", param->fasta_file);
			return qt;
		}
	}
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] == '>'){
			if(numseq && len >= param->minlen_cutoff && len <= param->maxlen_cutoff){
				if(len <= param->chop_at){
					tmp_seq[len] = 0;
				}else{
					len = param->chop_at;
					tmp_seq[len] = 0;
				}
				insert_hash_name(qt,param,tmp_seq,tmp_name,len,sn_len, count);
			}
			sn_len = 0;
			for(i = 1;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					tmp_name[sn_len] = 0;
					break;
				}else if(isspace((int)line[i])){
					tmp_name[sn_len]= '_';
				}else{
					tmp_name[sn_len] = line[i];
				}
				sn_len++;
			}
			numseq++;
			len = 0;
			count = 1.0;
			//count = atof(line+byg_end(param->ex_char, line));
			//fprintf(stderr,"%f	%s",count,line);
			
			/*for(i = strlen(line);i >= 0;i--){
				if(line[i] == param->ex_char[0]){
					count = atof(line+i+1);
					break;
				}
			}*/
			/*fprintf(stderr,"%f	%s\n",count,line);*/
			//if(!count){
			//	count = 1;
			//}
		}else{
			for(i = 0;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					tmp_seq[len] = 0;
					break;
				}
				tmp_seq[len] = toupper(line[i]);
				len++;
			}
		}
	}
	
	if(numseq && len >= param->minlen_cutoff && len <= param->maxlen_cutoff){
		if(len <= param->chop_at){
			tmp_seq[len] = 0;
		}else{
			len = param->chop_at;
			tmp_seq[len] = 0;
		}
		insert_hash_name(qt,param,tmp_seq,tmp_name,len,sn_len, count);
	}	
	
	fclose(file);
	return qt;
}

struct query_table* read_query_fasta_file(struct query_table* qt,struct parameters* param)
{
	FILE *file;
	char line[MAX_LINE];
	char tmp_seq[MAX_LINE];
	int i;
	int len = 0;
	double count = 0;
	int numseq = 0;
	
	if(!param->fasta_file){
		file = stdin;
		if (isatty(0)){
			//fprintf(stderr,"No STDIN..\n");
			return qt;
		}
	}else{
		if (!(file = fopen(param->fasta_file, "r" ))){
			fprintf(stderr,"Cannot open query file '%s'\n", param->fasta_file);
			return qt;
		}
	}
	
	while(fgets(line, MAX_LINE, file)){
		if(line[0] == '>'){
			if(numseq && len >= param->minlen_cutoff && len <= param->maxlen_cutoff){
				if(len <= param->chop_at){
					tmp_seq[len] = 0;
				}else{
					len = param->chop_at;
					tmp_seq[len] = 0;
				}
				//tmp_seq[len] = 0;
				insert_hash(qt,param,tmp_seq, len, count);
			}
			numseq++;
			len = 0;
			
			count = 1;
			//count = atof(line+byg_end(param->ex_char, line));
			//fprintf(stderr,"%f	%s",count,line);
			
			for(i = strlen(line);i >= 0;i--){
				if(line[i] == param->ex_char[0]){
					count = atof(line+i+1);
					break;
				}
			}
			/*count = atof(line+byg_end(param->ex_char, line));
			//fprintf(stderr,"%f	%s\n",count,line);
			if(!count){
				count = 1;
			}*/
		}else{
			for(i = 0;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					tmp_seq[len] = 0;
					break;
				}
				tmp_seq[len] = toupper(line[i]);
				len++;
			}
		}
	}
	
	if(numseq && len >= param->minlen_cutoff && len <= param->maxlen_cutoff){
		if(len <= param->chop_at){
			tmp_seq[len] = 0;
		}else{
			len = param->chop_at;
			tmp_seq[len] = 0;
		}
		//tmp_seq[len] = 0;
		insert_hash(qt,param,tmp_seq, len, count);
	}	
	
	fclose(file);
	return qt;
}




struct query_table* read_query_fasta_file_frag(struct query_table* qt,struct parameters* param,int frag)
{
	FILE *file;
	char line[MAX_LINE];
	char tmp_seq[MAX_LINE];
	char tmp_seq2[MAX_LINE];
	int i,j;
	int len = 0;
	double count = 0;
	int numseq = 0;
	
	if(!param->fasta_file){
		file = stdin;
		if (isatty(0)){
			//fprintf(stderr,"No STDIN..\n");
			return qt;
		}
	}else{
		if (!(file = fopen(param->fasta_file, "r" ))){
			fprintf(stderr,"Cannot open query file '%s'\n", param->fasta_file);
			return qt;
		}
	}
	
	while(fgets(line, MAX_LINE, file)){
		if(iscntrl((int)line[0])){
		}else if(line[0] == '>'){
			if(numseq && len >= param->minlen_cutoff && len <= param->maxlen_cutoff){
				//tmp_seq[len] = 0;
				if(len <= param->chop_at){
					tmp_seq[len] = 0;
				}else{
					len = param->chop_at;
					tmp_seq[len] = 0;
				}
				
				if(len < frag){ 
					//fprintf(stderr,"%s	%f\n",tmp_seq,count);
					insert_hash(qt,param,tmp_seq, len, count);
				}else{
					count = count / (double)(len+1-frag);
					for(i = 0; i < len+1 - frag;i++){
						for(j = 0;j < frag;j++){
							tmp_seq2[j] = tmp_seq[i+j];
						}
						tmp_seq2[j] = 0;
					//	fprintf(stderr,"%s	%f\n",tmp_seq2,count);
						insert_hash(qt,param,tmp_seq2, frag, count);
					}
				}
			}
			numseq++;
			len = 0;
			count = 1;
			//count = atof(line+byg_end(param->ex_char, line));
			//fprintf(stderr,"%f	%s",count,line);
			
			for(i = strlen(line);i >= 0;i--){
				if(line[i] == param->ex_char[0]){
					count = atof(line+i+1);
					break;
				}
			}
			/*count = atof(line+byg_end(param->ex_char, line));
			//fprintf(stderr,"%f	%s\n",count,line);
			if(!count){
				count = 1;
			}*/
		}else{
			for(i = 0;i < MAX_LINE;i++){
				if(iscntrl((int)line[i])){
					tmp_seq[len] = 0;
					break;
				}
				tmp_seq[len] = toupper(line[i]);
				len++;
			}
		}
	}
	
	if(numseq && len >=param->minlen_cutoff && len <= param->maxlen_cutoff){
		//tmp_seq[len] = 0;
		if(len <= param->chop_at){
			tmp_seq[len] = 0;
		}else{
			len = param->chop_at;
			tmp_seq[len] = 0;
		}
		if(len < frag){ 
			insert_hash(qt,param,tmp_seq, len, count);
		}else{
			count = count / (double)(len+1-frag);
			for(i = 0; i < len+1 -frag;i++){
				for(j = 0;j < frag;j++){
					tmp_seq2[j] = tmp_seq[i+j];
				}
				tmp_seq2[j] = 0;
				insert_hash(qt,param,tmp_seq2, frag, count);
			}
		}
	}	
	
	fclose(file);
	return qt;
}



struct fasta* read_bin_aln(struct fasta* p,const char* file)
{
	int fd = 0;
	int i = 0;
	int test;
	int* len_array = 0;
	int mer_hash = 0;
	int version = 0;
	
	if ((fd = open(file, O_RDONLY )) == -1){
		fprintf(stderr,"Cannot open file %s\n",file);
		exit(-1);
	}
	
	test = read(fd,&version,sizeof(int));
	assert(test >= 0);
	
	if(version != SUFFIX_VERSION){
		fprintf(stderr,"Warning: There has been a change to the format of the suffix files in the recent version. You need to re-build all suffix files.\n");
		exit(-1);
		free(p);
	}
	p = malloc(sizeof(struct fasta));
	assert(p != 0);
	
	test = read(fd,&mer_hash,sizeof(int));
	assert(test>=0);
	
	test = read(fd,&p->string_len,sizeof(int));
	assert(test >= 0);
	test = read(fd,&p->numseq,sizeof(int));
	assert(test >= 0);
	test = read(fd,&p->max_len,sizeof(int));
	assert(test >= 0);
	
	p->sn= malloc(sizeof(char*)*p->numseq);	
	assert(p->sn != 0);
	p->s_index= malloc(sizeof(int)*(p->numseq+1));	
	assert(p->s_index != 0);
	len_array = malloc(sizeof(int)*p->numseq); 
	assert(len_array != 0);
	test = read(fd,len_array,sizeof(int)*p->numseq);
	assert(test >= 0);
	
	for(i = 0; i < p->numseq;i++){
		p->sn[i] = malloc(sizeof(char)*(len_array[i]+1));
		assert(p->sn[i] != 0);
		test = read(fd,p->sn[i],sizeof(char)*len_array[i]);
		assert(test >= 0);
		p->sn[i][len_array[i]] = 0;
	}
	test = read(fd,p->s_index,sizeof(int)*(p->numseq+1));
	assert(test >= 0);

 	p->string = malloc(sizeof(char)*(p->string_len+1+17));
	assert(p->string != 0);

	test = read(fd,p->string,sizeof(char)*p->string_len+17);
	assert(test >= 0);
	p->string[p->string_len+17] = 0;
 	p->suffix = malloc(sizeof(int)*(p->string_len+1));
	assert(p->suffix != 0);
	test = read(fd,p->suffix,sizeof(int)*(p->string_len+1));
	assert(test >= 0);
	
	if(mer_hash){
		p->mer_hash = malloc(sizeof(int)*(134217728));
		assert(p->mer_hash != 0);
		test = read(fd,p->mer_hash,sizeof(int)*(134217728));
		assert(test >= 0);
	}else{
		p->mer_hash= 0;
	}
	
	close(fd);
	free(len_array);
	return p;
}






