/*
	misc.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2006 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"

void mer_hash(struct fasta* target,int len)
{
	int i;
	unsigned int t = 0;
	//unsigned int prev = 0;
	//unsigned int covered = 0;
	//unsigned long int numseq = 4; 

	for(i = 0; i < target->string_len;i++){
		t = tag_index(target->string + target->suffix[i]);
		target->mer_hash[t >> 5] |= (1 << (t & 0x1F));
		
		/*if(j > i){
		fprintf(stderr," %d	%d	%d\n",i,j,j-i);
		}*/
		//if(t != prev){
		//	covered++; 
		//}
		//prev = t;
		//fprintf(stderr,"%x\n",t);
		//t = t >> 12;
		//t = t << 1;
		//fprintf(stderr,"%x\n",t);
		//if(target->boost[t] > i-1){
		//	target->boost[t] = i-1;
		//}
		//target->boost[t+1] = i+1;
		
		
		
	}
	
	//for(i = 0;i < len-1;i++){
	//	numseq *= 4;
	//}
	//fprintf(stderr,"LEN:%d	%ud	/ %lud = %f\n",len,covered,numseq,(double)covered/(double)numseq*100.0);
}

unsigned int tag_index(const char* tag)
{
	unsigned int t = 0;
	//int len =16;
	//int c = len*2-1;
	int c = 31;//
	int j;
	for(j = 0; j < 16;j++){
		switch (tag[j]){
			case 'A':
			case '0':
				c--;
				break;
			case 'C':
			case '1':
				c--;
				t += 1 << c;
				break;
			case 'G':
			case '2':
				t += 1 << c;
				c--;
				break;
			case 'T':
			case '3':
				c--;
				t+= 3 << c;
				break;
		}
		c--;
	}
	return t;
}

unsigned int tag_index2(const char* tag)
{
	unsigned int t = 0;
	//const char* rev = "0B1DEF2HIJKLMNOPQRS3UVWXYZ";
	const int rev[26] = {0,0,1,0,0,0,2,0,0,0,0,0,0,0,0,0,0,0,0,3,0,0,0,0,0,0};
	//int len =16;
	//int c = len*2-1;
	//int c = 31;//
	int j;
	for(j = 0; j < 16;j++){
		t = t << 2;
		t |= rev[tag[j]-65];
	}
	return t;
}
unsigned int tag_index_solid(const char* tag)
{
	unsigned int t = 0;
	int j;
	for(j = 0; j < 16;j++){
		t = t << 2;
		t |= tag[j]-48;
	}
	return t;
}

int test_tag_index(const char* tag,const int* mer_hash)
{
	unsigned int t = 0;
	int len =16;
	int c = len*2-1;
	int j;
	for(j = 0; j < len;j++){
		switch (tag[j]){
			case 'A':
			case '0':
				c--;
				break;
			case 'C':
			case '1':
				c--;
				t += 1 << c;
				break;
			case 'G':
			case '2':
				t += 1 << c;
				c--;
				break;
			case 'T':
			case '3':
				c--;
				t+= 3 << c;
				break;
		}
		c--;
	}
	
	if(!mer_hash[t>>5] & (1<<(t & 0x1F))){
		return 0;
	}

	while(j < strlen(tag)){
		t = t <<2 ;
		switch (tag[j]){
			case 'A':
			case '0':
				break;
			case 'C':
			case '1':
				t += 1;
				break;
			case 'G':
			case '2':
				t += 2;
				break;
			case 'T':
			case '3':
				t+= 3;
				break;
		}	
		if(!mer_hash[t>>5] & (1<<(t & 0x1F))){
			return 0;
		}
		j++;
	}
	return 1;
}

void pop_stats(struct query_table* qt,struct parameters* param)
{
	int i,j;
	unsigned int* tmp = 0;
	double c = 0.0;
	
	
	if(!param->upper_limit){
		for(i = 0; i < qt->size;i++){
			tmp = qt->park[i]->stats;
			c = qt->park[i]->count;
			/*for(j = 0;j < 6;j++){
				fprintf(stderr,"%d	",qt->park[i]->stats[j]);
			}
			fprintf(stderr,"\n");*/
			for(j = 2;j < 6;j++){
				tmp[j] += tmp[j-1];
			}
			for(j = 0; j < 6;j++){
				if(tmp[j] == 0){
					qt->stats[j][0] += c;
				}else if(tmp[j] == 1){
					qt->stats[j][1] += c;
				}else if(tmp[j] <=10){
					qt->stats[j][2] += c;
				}else if(tmp[j] <=100){
					qt->stats[j][3] += c;
				}else if(tmp[j] > 100){
					qt->stats[j][4] += c;
				}
			}
		}
	}else if(param->upper_limit == 1){
		for(i = 0; i < qt->size;i++){
			tmp = qt->park[i]->stats;
			c = qt->park[i]->count;
			/*for(j = 0;j < 6;j++){
			fprintf(stderr,"%d	",qt->park[i]->stats[j]);
		}
			fprintf(stderr,"\n");*/
			for(j = 2;j < 6;j++){
				tmp[j] += tmp[j-1];
			}
			j = 0;
			//for(j = 0; j < 6;j++){
				if(tmp[j] == 0){
					qt->stats[j][0] += c;
				}else if(tmp[j] == 1){
					qt->stats[j][1] += c;
				}else if(tmp[j] > 1){
					qt->stats[j][2] += c;
				}
			//}
		}
	}else{
		for(i = 0; i < qt->size;i++){
			tmp = qt->park[i]->stats;
			c = qt->park[i]->count;
			/*for(j = 0;j < 6;j++){
			fprintf(stderr,"%d	",qt->park[i]->stats[j]);
		}
			fprintf(stderr,"\n");*/
			for(j = 2;j < 6;j++){
				tmp[j] += tmp[j-1];
			}
			j = 0;
			//for(j = 0; j < 6;j++){
				if(tmp[j] == 0){
					qt->stats[j][0] += c;
				}else if(tmp[j] == 1){
					qt->stats[j][1] += c;
				}else if(tmp[j] <= param->upper_limit){
					qt->stats[j][2] += c;
				}else if(tmp[j] > param->upper_limit){
					qt->stats[j][3] += c;
				}
			//}
		}
	}
}

void check_sequences(struct query_table* qt,struct parameters* param)
{
	int allowed[256];
	
	int i,j,c;
	int is_solid = 0;
	int is_short_seq = 0;
	
	for(i =0; i < 256;i++){
		allowed[i] = 0;
	}
	
	allowed[48] = 1;
	allowed[49] = 1;
	allowed[50] = 1;
	allowed[51] = 1;
	
	allowed[65] = 1;
	allowed[67] = 1;
	allowed[71] = 1;
	allowed[84] = 1;
	allowed[78] = 1;
	
	allowed[88] = 1;
	
	/*allowed[97] = 1;
	allowed[99] = 1;
	allowed[103] = 1;
	allowed[116] = 1;
	allowed[110] = 1;
	allowed[120] = 1;*/
	
	allowed[46] = 1;
	
	if(!param->quiet_flag){
		fprintf(stderr,"\n	Checking for presence of valid sequence characters (allowed:\"");
		for(i = 0; i <255;i++){
			if(allowed[i]){
				fprintf(stderr,"%c",(char)i);
			}
		}
		fprintf(stderr,"\")\n");
	}
	c = 0;
	for (i = 0; i < qt->size;i++){
		if(qt->park[i]->len < 16){
			is_short_seq++;
		}
		for(j = 0;j < qt->park[i]->len;j++){
			if(!allowed[(int)qt->park[i]->seq[j]]){
				fprintf(stderr,"	WARNING:Sequence: %s not valid setting count to 0\n",qt->park[i]->seq);
				qt->park[i]->count = 0.0;
				c++;
				break;
			}
		}
	}
	
	if(!param->short_seq && is_short_seq){
		if(!param->quiet_flag){
			fprintf(stderr,"	Detected %d short sequences - switching to short mode.\n",is_short_seq);
		}
		param->short_seq = is_short_seq;
	}
	if(!param->solid){
		for(i= 0; i < qt->size;i++){
			if(isdigit(qt->park[i]->seq[1])){
				is_solid++;
			}
		}
		if(is_solid){
			param->solid = is_solid;
			if(!param->quiet_flag){
				fprintf(stderr,"	Found %d color space sequences - switching to color space alignment.\n",is_solid);
			}
		}
	}
	if(param->solid && param->short_seq){
		fprintf(stderr,"	WARNING: currently nexalign only support alignment with color space sequences longer than 16 colors.\n");
		exit(-1);
	}
	//fprintf(stderr,"Short:%d	Solid:%d\n",param->short_seq,param->solid);
	
	if(c){
		fprintf(stderr,"	WARNING:%d invalid sequences\n\n",c);
	}else{
		if(!param->quiet_flag){
		fprintf(stderr,"	All sequences ok (%d unique,%1.0f total).\n\n",qt->numseq,qt->total_expression);
		}
	}
}

void pop_ann_tree(struct query_table* qt,int max_no_ann_overlap)
{
	int i;
	for(i= 0; i < qt->size;i++){
		if(pop(qt->park[i]->stats[6]) <= max_no_ann_overlap){
			qt->tree = insert(qt->tree,qt->park[i]->stats[6],qt->park[i]->count,&qt->tree_size);
		}else{
			qt->tree = insert(qt->tree,-1,qt->park[i]->count,&qt->tree_size);
		}		
	}
}

void clear_counts(struct query_table* qt)
{
	int i,j;
	for(i= 0; i < qt->size;i++){
		for(j =0;j < 8;j++){
			qt->park[i]->stats[j] = 0;
		}
	}
}

void clear_6_counts(struct query_table* qt)
{
	int i,j;
	for(i= 0; i < qt->size;i++){
		for(j =0;j < 6;j++){
			qt->park[i]->stats[j] = 0;
		}
	}
}

void set_class(struct query_table* qt,struct parameters* param, int db)
{
	int i,j; 
	int type = -1;
	char tmp[MAX_LINE];
	tmp[0] = '_';
	for(i = 0;i < param->num_types;i++){
		for(j = 0; j <= strlen(param->types[i]);j++){
			tmp[j+1] = param->types[i][j];
		}
		if(is_substr(tmp,param->infile[db])){
			type = i;
			break;
		}
	}
	if(type == -1){
		if (param->type_file){
			fprintf(stderr,"WARNING:%s could not be associated to a annotation type in file\n",param->type_file);
		}
		type = db;
	}else{
		if(!param->quiet_flag){
			fprintf(stderr,"%s is %s type\n",param->infile[db],param->types[type]);
		}
	}
		
	for(i= 0; i < qt->size;i++){
		if(qt->park[i]->stats[0]){// > qt->park[i]->stats[7]){
			qt->park[i]->stats[6]  |= 1<< type;
		}
		qt->park[i]->stats[7] += qt->park[i]->stats[0];
	}
}

int test_solid(struct query_table* qt)
{
	int solid = 0;
	int i;
	for(i= 0; i < qt->size;i++){
		if(isdigit(qt->park[i]->seq[1])){
			solid++;
		} 
	}
	return solid;
}

int arbins(const int* x, const int size, const int y)
{
	int h,l,m;
	l = 0;
	h = size+1;
	while(h-l > 1){
		m = (l+h)/2;
		if(x[m] > y){
			h = m;
		}else{
			l = m;
		}
	}
	return l;
}

char* shorten_name(char* p)
{
	char* n = 0;
	int i,j;
	j = 0;
	for(i = 0; i < strlen(p);i++){
		if(p[i] == '/'){
			j = i;
		}
	}
	if(j != 0){
		n = p+j+1;
	}else{
		n = p;
	}
	
	return n;
}

void reverse_complement(const char* p,char* rc,int len)
{
	int i,c;
	c = 0;
	//rc = malloc(sizeof(char)*(len+1));
	for(i =len-1; i >= 0;i--){
		switch(p[i]){
			case 'A':
				rc[c] = 'T';
				break;
			case 'C':
				rc[c] = 'G';
				break;
			case 'G':
				rc[c] = 'C';
				break;
			case 'T':
				rc[c] = 'A';
				break;
			default:
				rc[c] = p[i];
				break;
		}
		c++;
	}
	rc[c] = 0;
}

void reverse_complement2(const char* p,char* rc,int len)
{
	const char* rev = "TBGDEFCHIJKLMNOPQRSAUVWXYZ"; 

	int i,c;
	c = 0;
	//rc = malloc(sizeof(char)*(len+1));
	for(i =len-1; i >= 0;i--){
		rc[c] = rev[p[i]-65];
		/*switch(p[i]){
			case 'A':
				rc[c] = 'T';
				break;
			case 'C':
				rc[c] = 'G';
				break;
			case 'G':
				rc[c] = 'C';
				break;
			case 'T':
				rc[c] = 'A';
				break;
			default:
				rc[c] = p[i];
				break;
		}*/
		c++;
	}
	rc[c] = 0;
}

int base_space_errors(const char* p,const char* t,int len)
{
	char* alphabet = "ACGT";
	int bs_code[4][4] = {
	{0,1,2,3},
 	{1,0,3,2},
 	{2,3,0,1},
 	{3,2,1,0}
	};
	int bse = 0;
	int base =0;
	int i;
	//if(p[0] != t[0]){
	//	bse++;
	//}
	//fprintf(stderr,"%c	%c\n",p[0],t[0]);
	switch(p[0]){
		case 'A':
			base = 0;
			break;
		case 'C':
			base = 1;
			break;
		case 'G':
			base = 2;
			break;
		case 'T':
			base = 3;
			break;
	}
	
	for(i = 1; i < len;i++){
		base = bs_code[base][p[i]-48];
	//	fprintf(stderr,"%c	%c	%d\n",alphabet[base],t[i],bse);
		if(alphabet[base] != t[i]){
			bse++;
		}
	}
	return bse;
}

int base_space_errors_rev(const char* p,const char* t,int len)
{
	char* rc = 0;
	rc = malloc(sizeof(char)*(len+1));
	assert(rc != 0);
	reverse_complement(t,rc,len);
	
	char* alphabet = "ACGT";
	int bs_code[4][4] = {
		{0,1,2,3},
  		{1,0,3,2},
		{2,3,0,1},
		{3,2,1,0}
	};
	int bse = 0;
	int base =0;
	int i;
	if(p[0] != rc[0]){
		bse++;
	}
	//fprintf(stderr,"%c	%c\n",p[0],rc[0]);
	switch(p[0]){
		case 'A':
			base = 0;
			break;
		case 'C':
			base = 1;
			break;
		case 'G':
			base = 2;
			break;
		case 'T':
			base = 3;
			break;
	}
	if(strlen(t) < len){
		bse += len - strlen(t);
		len = strlen(t);
	}
	for(i = 1; i < len;i++){
		base = bs_code[base][p[i]-48];
	//	fprintf(stderr,"%c	%c	%d\n",alphabet[base],rc[i],bse);
		if(alphabet[base] != rc[i]){
			bse++;
		}
	}
	free(rc);
	return bse;
}

void reverse(const char* p,char* rc,const int len)
{
	register int i;//,c;
	//c = 0;
	//rc = malloc(sizeof(char)*(len+1));
	for(i =len-1; i >= 0;i--){
		*rc = p[i];
		rc++;
	}
	//rc[c] = 0;
	//return rc;
}

int pop(int x)
{
	int n = 0;
	while(x != 0){
		n = n +1;
		x = x &(x-1);
	}
	return n;
}

