/*
	string_matching.c
	
	Released under GPL - see the 'COPYING' file   
	
	Copyright (C) 2007 Timo Lassmann <timolassmann@gmail.com>
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    
	Please send bug reports, comments etc. to:
	timolassmann@gmail.com
*/

#include "ngs_toolkit.h"

int binsearch_down4(const char*p,const char* text,const int* suffix,const int len,int l,int h)
{
	int m = 0;
	if (t_long_strncmp(p,text+suffix[l],len)<= 0){
		return l;
	}else
	if(t_long_strncmp(p,text+suffix[h],len) >  0){
		return h;
	}else{
		while(h-l > 1){
			m = (l+h)/2;
			if(t_long_strncmp(p,text+suffix[m],len) <= 0){
				h = m;
			}else{
				l = m;
			}
		}
	}
	return l;
}

int binsearch_up4(const char*p,const char* text,const int* suffix,const int len,int l,int h)
{
	int m = 0;
	if (t_long_strncmp(p,text+suffix[l],len) <= 0){
		return l;
	}else if(t_long_strncmp(p,text+suffix[h],len) >  0){
		//fprintf(stderr,"Ping\n");
		return h;
	}else{
		while(h-l > 1){
			m = (l+h)/2;
			if(t_long_strncmp(p,text+suffix[m],len) < 0){
				h = m;
			}else{
				l = m;
			}
		}
	}
	return l;
}

int binsearch_down(const char*p,const char* text,const int* suffix,const int len,int h)
{
	int m = 0;
	int l = 0;
	/*if (t_long_strncmp(p,text+suffix[l],len)<= 0){
		l = l;
	}else */
	if(t_long_strncmp(p,text+suffix[h],len) >  0){
		return h;
	}else{
		while(h-l > 1){
			//m = (l+h)/2;
			m = (l + h) >> 1;
			if(t_long_strncmp(p,text+suffix[m],len) <= 0){
				h = m;
			}else{
				l = m;
			}
		}
	}
	return l;
}

int binsearch_up(const char*p,const char* text,const int* suffix,const int len,int h)
{
	int m = 0;
	int l = 0;
	/*if (t_long_strncmp(p,text+suffix[l],len)<= 0){
		l = l;
	}else*/
	if(t_long_strncmp(p,text+suffix[h],len) >  0){
		return h;
	}else{
		while(h-l > 1){
			//m = (l+h)/2;
			m = (l + h) >> 1;
			if(t_long_strncmp(p,text+suffix[m],len) < 0){
				h = m;
			}else{
				l = m;
			}
		}
	}
	return l;
}

void binsearch_up3(struct suffix_pos* sp,const char*p,const char* text,const int* suffix,const int len,int h)
{
	int m = 0;
	int l = 0;
	/*if(t_long_strncmp(p,text+suffix[h],len) >  0){
		//return h;
		//l =0;
		//sp->e_p = -1;
	}else{*/
		while(h-l > 1){
			m = (l+h)/2;
			if(t_long_strncmp(p,text+suffix[m],len) < 0){
				h = m;
			}else{
				l = m;
			}
		}
	//}
	if(!t_long_strncmp(p,text+suffix[l],len)){
		sp->e_p = l;
		m = 0;
		h = l;
		l = 0;
		/*if(t_long_strncmp(p,text+suffix[h],len) >  0){
			fprintf(stderr,"possible.....");
			//return h;
		}else{*/
			while(h-l > 1){
				m = (l+h)/2;
				if(t_long_strncmp(p,text+suffix[m],len) <= 0){
					h = m;
				}else{
					l = m;
				}
			}
		//}
		sp->s_p = l;
	}else{
		sp->e_p = 0;
		sp->s_p = 0;
	}
}


int binsearch_up2(const char*p,const char* text,const int* suffix,const int len,int h)
{
	int l = -1;
	if(t_long_strncmp(p,text+suffix[134217727],len) >= 0){
		l = h-134217728;
	}
	if(t_long_strncmp(p,text+suffix[l+67108864],len) >= 0){
		l+= 67108864;
	}
	if(t_long_strncmp(p,text+suffix[l+33554432],len) >= 0){
		l+= 33554432;
	}
	if(t_long_strncmp(p,text+suffix[l+16777216],len) >= 0){
		l+= 16777216;
	}
	if(t_long_strncmp(p,text+suffix[l+8388608],len) >= 0){
		l+= 8388608;
	}
	if(t_long_strncmp(p,text+suffix[l+4194304],len) >= 0){
		l+= 4194304;
	}
	if(t_long_strncmp(p,text+suffix[l+2097152],len) >= 0){
		l+= 2097152;
	}
	if(t_long_strncmp(p,text+suffix[l+1048576],len) >= 0){
		l+= 1048576;
	}
	if(t_long_strncmp(p,text+suffix[l+524288],len) >= 0){
		l+= 524288;
	}
	if(t_long_strncmp(p,text+suffix[l+262144],len) >= 0){
		l+= 262144;
	}
	if(t_long_strncmp(p,text+suffix[l+131072],len) >= 0){
		l+= 131072;
	}
	if(t_long_strncmp(p,text+suffix[l+65536],len) >= 0){
		l+= 65536;
	}
	if(t_long_strncmp(p,text+suffix[l+32768],len) >= 0){
		l+= 32768;
	}
	if(t_long_strncmp(p,text+suffix[l+16384],len) >= 0){
		l+= 16384;
	}
	if(t_long_strncmp(p,text+suffix[l+8192],len) >= 0){
		l+= 8192;
	}
	if(t_long_strncmp(p,text+suffix[l+4096],len) >= 0){
		l+= 4096;
	}
	if(t_long_strncmp(p,text+suffix[l+2048],len) >= 0){
		l+= 2048;
	}
	if(t_long_strncmp(p,text+suffix[l+1024],len) >= 0){
		l+= 1024;
	}
	if(t_long_strncmp(p,text+suffix[l+512],len) >= 0){
		l+= 512;
	}
	if(t_long_strncmp(p,text+suffix[l+256],len) >= 0){
		l+= 256;
	}
	if(t_long_strncmp(p,text+suffix[l+128],len) >= 0){
		l+= 128;
	}
	if(t_long_strncmp(p,text+suffix[l+64],len) >= 0){
		l+= 64;
	}
	if(t_long_strncmp(p,text+suffix[l+32],len) >= 0){
		l+= 32;
	}
	if(t_long_strncmp(p,text+suffix[l+16],len) >= 0){
		l+= 16;
	}
	if(t_long_strncmp(p,text+suffix[l+8],len) >= 0){
		l+= 8;
	}
	if(t_long_strncmp(p,text+suffix[l+4],len) >= 0){
		l+= 4;
	}
	if(t_long_strncmp(p,text+suffix[l+2],len) >= 0){
		l+= 2;
	}
	if(t_long_strncmp(p,text+suffix[l+1],len) >= 0){
		l+= 1;
	}
	
	//if(l > h || t_long_strncmp(p,text+suffix[l],len)!= 0){
	//	l = -1;
	//}
	return l;
}


int is_substr(const char* pattern,const char*text)
{
	const char* tmp = 0;
	int i  = 0;
	int m = strlen(pattern);
	int n = strlen (text);
	if (m > n){
		i = m;
		m = n;
		n = i;
		tmp = text;
		text = pattern;
		pattern = tmp;
	}
	
	for(i = 0; i <= (n - m);i++){
		if(!strncmp(text+i,pattern,m)){
			return 1;
		}
	}
	return 0;
}


int byg_end(const char* pattern,const char*text)
{
	const char* tmp = 0;
	int Tc;
	int i  = 0;
	int s = 0;
	int T[256];
	for (i = 0;i < 256;i++){ 
		T[i] = 0; 
	}
	
	int m = strlen(pattern);
	int n = strlen (text);
	if (m > n){
		i = m;
		m = n;
		n = i;
		tmp = text;
		text = pattern;
		pattern = tmp;
	}
	
	int mb = (1 << (m-1));
	
	for (i= 0;i < m;i++){
		T[(int)pattern[i]] |= (1 << i);
	}
	
	for (i = 0;i < n;i++){
		s <<= 1;
		s |= 1;
		if(!text[i]){
			return -1;
		}
		Tc = T[(int)text[i]];
		s &= Tc;
		if(s & mb){
			return i+1;
		}
	}
	return 0;
}


long int t_long_strncmp(const char *s,const char *t,const int len)
{
	long unsigned int* a = (long unsigned int*) s;
	long unsigned int* b = (long unsigned int*) t;
	register int i;
	int interval = len / sizeof(long unsigned int);
	/*fprintf(stderr,"%c%c%c%c%c%c%c%c\n",s[0],s[1],s[2],s[3],s[4],s[5],s[6],s[7]);
	fprintf(stderr,"%lud\n",a[0]);
	a[0] = rev(a[0]);
	fprintf(stderr,"%lud\n",a[0]);*/
	//register long unsigned int x = 0;
	//register long unsigned int y = 0;
	for (i = 0; i < interval;i++){
		if(a[i] != b[i]){
			return rev(a[i]) - rev(b[i]);
			//x = rev(a[i]);
			//y = rev(b[i]);
			//return x - y;
		}
	}
	for (i = interval*sizeof(long int);i < len; i++){
		if(s[i] != t[i]){
 			return s[i] - t[i];
		}
	}
	return 0;
}

#if defined(__alpha__)||defined(__ia64__)||defined(__ppc64__)||defined(__s390x__)||defined(__x86_64__)
unsigned long int rev(unsigned long int x)
{
	x = (x & 0x00FF00FF00FF00FF) << 8 | (x & 0xFF00FF00FF00FF00) >> 8;
	x = (x & 0x0000FFFF0000FFFF) << 16 | (x  & 0xFFFF0000FFFF0000) >> 16;
	x = (x & 0x00000000FFFFFFFF) << 32 | (x  & 0xFFFFFFFF00000000) >> 32;
	return x;
}
#else
unsigned long int rev(unsigned long int x)
{
	x = (x & 0x00FF00FF) << 8 | (x & 0xFF00FF00) >> 8;
	x = (x & 0x0000FFFF) << 16 | (x  & 0xFFFF0000) >> 16;
	return x;
}
#endif



